# pylint: skip-file

"""Testing"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cxlint import CxLint

if __name__ == "__main__":
    agent_local_path = "data/agent"
    agent_id = 'projects/prebuilt-components-qa/locations/global/agents/872bd829-3ff6-4a3f-a242-cc5fa769a56c'
    gcs_path = 'gs://pmarlow_dev_agents/demo_agent_json.zip'

    naming_conventions = {
        "agent_name": ".*",
        "flow_name": ".*",
        "intent_head_name": "head_intent.*",
        "intent_confirmation_name": ".*",
        "intent_escalation_name": ".*",
        "intent_generic_name": ".*",
        "entity_type_name": ".*",
        "page_generic_name": ".*",
        "page_with_form_name": ".*",
        "page_with_webhook_name": ".*",
        "test_case_name": ".*",
        "webhook_name": ".*"
        }

    cxlint = CxLint(
        agent_id=agent_id,
        naming_conventions=naming_conventions,
        load_gcs=True,
        # agent_type='chat',
        # language_code=['en'],
        # resource_filter=["flows"],
        # flow_include_list=['Steering'],
        # intent_include_pattern='sup'
        output_file="/Users/pmarlow/eng/cxlint/data/logs.txt",
    )

    agent_file = cxlint.gcs.download_gcs(gcs_path, agent_local_path)
    cxlint.gcs.unzip(agent_file, agent_local_path)

    cxlint.lint_agent(agent_local_path)



"""Common methods and helper functions used throughout library."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re

from configparser import ConfigParser
from typing import Dict, List, Union
from cxlint.resources.types import Intent, EntityType

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
)


class Common:
    """Common methods and helper functions used throughout library."""

    @staticmethod
    def load_naming_conventions(config: ConfigParser) -> Dict[str, str]:
        """Loads the Naming Convention styles into a map."""
        data = dict(config.items("NAMING CONVENTIONS"))

        for key, value in data.items():
            if value == "":
                data[key] = None

        return data

    @staticmethod
    def load_message_controls(config: ConfigParser) -> Dict[str, str]:
        """Loads the config file for message control into a map."""
        msg_list = (
            config["MESSAGES CONTROL"]["disable"].replace("\n", "").split(",")
        )

        msg_dict = {msg: False for msg in msg_list}

        return msg_dict

    @staticmethod
    def load_agent_type(config: ConfigParser) -> Dict[str, str]:
        """Loads the config file for agent type."""
        agent_type = config["AGENT TYPE"]["type"]

        return agent_type

    @staticmethod
    def load_resource_filter(config: ConfigParser) -> List[str]:
        """Loads the config file for agent resource filtering."""
        resource_filter = (
            config["AGENT RESOURCES"]["include"].replace("\n", "").split(",")
        )

        resource_dict = {
            "entity_types": True,
            "flows": True,
            "intents": True,
            "test_cases": True,
            "webhooks": True,
        }

        if len(resource_filter) == 1 and resource_filter[0] == "":
            resource_filter = None

        if resource_filter:
            for resource in resource_dict:
                if resource not in resource_filter:
                    resource_dict[resource] = False

        return resource_dict

    @staticmethod
    def load_agent_id(config: ConfigParser) -> str:
        """Loads the Agent ID from the config file if provided."""
        agent_id = config["AGENT ID"]["id"]

        return agent_id

    @staticmethod
    def calculate_rating(total_issues: int, total_inspected: int) -> float:
        """Calculate the final rating for the linter stats."""
        if total_inspected > 0:
            rating = (1 - (total_issues / total_inspected)) * 10

        else:
            rating = 10

        return rating

    @staticmethod
    def parse_filepath(in_path: str, resource_type: str) -> str:
        """Parse file path to provide quick reference for linter log."""

        regex_map = {
            "flow": r".*\/flows\/([^\/]*)",
            "page": r".*\/pages\/([^\/]*)\.",
            "entity_type": r".*\/entityTypes\/([^\/]*)",
            "intent": r".*\/intents\/([^\/]*)",
            "route_group": r".*\/transitionRouteGroups\/([^\/]*)",
            "webhook": r".*\/webhooks\/([^\/]*)\."
        }
        resource_name = re.match(regex_map[resource_type], in_path).groups()[0]

        return resource_name

    @staticmethod
    def clean_display_name(display_name: str):
        """Replace cspecial haracters from map for the given display name."""
        patterns = {
            "%22": '"',
            "%23": "#",
            "%24": "$",
            "%26": "&",
            "%27": "'",
            "%28": "(",
            "%29": ")",
            "%2c": ",",
            "%2f": "/",
            "%3a": ":",
            "%3c": "<",
            "%3d": "=",
            "%3e": ">",
            "%3f": "?",
            "%5b": "[",
            "%5d": "]",
            "%e2%80%9c": "“",
            "%e2%80%9d": "”",
        }

        for key, value in patterns.items():
            if key in display_name:
                display_name = display_name.replace(key, value)

        return display_name

    @staticmethod
    def load_lang_code_filter(config: ConfigParser) -> str:
        """Loads the language code filter for Intent Training Phrases."""
        lang_codes = config["INTENTS"]["language_code"]
        lang_code_list = lang_codes.split(",")

        if len(lang_code_list) == 1 and lang_code_list[0] == "":
            lang_code_list = None

        return lang_code_list

    @staticmethod
    def get_file_based_on_lang_code_filter(
        resource: Union[Intent, EntityType], lang_code, lang_code_filter
    ) -> Union[Intent, EntityType]:
        """Gets the file if it qualifies for lang_code filter."""
        # TODO pmarlow: Refactor for better readability

        if lang_code_filter:
            if lang_code in lang_code_filter:
                if isinstance(resource, Intent):
                    filename = resource.training_phrases[lang_code]["file_path"]

                if isinstance(resource, EntityType):
                    filename = resource.entities[lang_code]["file_path"]
            else:
                filename = None

        else:
            if isinstance(resource, Intent):
                filename = resource.training_phrases[lang_code]["file_path"]

            if isinstance(resource, EntityType):
                filename = resource.entities[lang_code]["file_path"]

        return filename

    @staticmethod
    def resource_precheck(
        agent_local_path: str,
        resource_filter: Dict[str, bool]):
        """PreLint Check to ensure the resource directory exists.

        The `resources` dict uses camelCase because the file structure is in
        camelCase. The `resource_filter` uses snake_case because the incoming
        .cxlintrc file stores the data this way. There is some minor string
        manipulation for resources that have 2 names to ensure we're checking
        everything appropriately."""
        resources = {
            "agents": False,
            "flows": False,
            "entityTypes": False,
            "intents": False,
            "testCases": False,
            "webhooks": False
        }

        if "agent.json" in os.listdir(agent_local_path):
            resources["agents"] = True

        # Ensure resource directories exist
        for resource, _ in resources.items():

            # skip cause we handle this case outside loop
            if resource == "agents":
                pass

            elif resource in os.listdir(agent_local_path):
                resources[resource] = True

        # Clean up dict so we can use snake_case from here on
        resources["entity_types"] = resources["entityTypes"]
        resources["test_cases"] = resources["testCases"]
        resources.pop("entityTypes", None)
        resources.pop("testCases", None)

        # Check against user requrest filters
        for resource, value in resource_filter.items():
            if not value:
                resources[resource] = False

        return resources



"""Intent linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from typing import Dict
from configparser import ConfigParser

from cxlint.common import Common
from cxlint.rules.intents import IntentRules
from cxlint.resources.types import Intent, LintStats


class Intents:
    """Intent linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.agent_id = Common.load_agent_id(config)
        self.disable_map = Common.load_message_controls(config)
        self.lang_code_filter = Common.load_lang_code_filter(config)
        self.naming_conventions = Common.load_naming_conventions(config)

        self.include_filter = self.load_include_filter(config)
        self.exclude_filter = self.load_exclude_filter(config)

        self.rules = IntentRules(console, self.disable_map)

    @staticmethod
    def load_naming_conventions(intent: Intent, styles: Dict[str, str]):
        """Load all Intent naming conventions to the current object."""
        intent.naming_pattern_generic = styles.get("intent_generic_name")
        intent.naming_pattern_head = styles.get("intent_head_name", None)
        intent.naming_pattern_confirmation = styles.get(
            "intent_confirmation_name")
        intent.naming_pattern_escalation = styles.get(
            "intent_escalation_name", None)

        return intent

    @staticmethod
    def load_include_filter(config: ConfigParser) -> str:
        """Loads the include pattern for Intent display names."""
        pattern = config["INTENTS"]["include"]

        return pattern

    @staticmethod
    def load_exclude_filter(config: ConfigParser) -> str:
        """Loads the exclude pattern for Intent display names."""
        pattern = config["INTENTS"]["exclude"]

        if pattern == "":
            pattern = None

        return pattern

    @staticmethod
    def parse_lang_code(lang_code_path: str) -> str:
        """Extract the language_code from the given file path."""

        first_parse = lang_code_path.split("/")[-1]
        lang_code = first_parse.split(".")[0]

        return lang_code

    @staticmethod
    def build_lang_code_paths(intent: Intent):
        """Builds dict of lang codes and file locations.

        The language_codes and paths for each file are stored in a dictionary
        inside of the Intent dataclass. This dict is access later to lint each
        file and provide reporting based on each language code.
        """
        root_dir = intent.dir_path + "/trainingPhrases"

        for lang_file in os.listdir(root_dir):
            lang_code = lang_file.split(".")[0]
            lang_code_path = f"{root_dir}/{lang_file}"
            intent.training_phrases[lang_code] = {"file_path": lang_code_path}

    @staticmethod
    def build_intent_path_list(agent_local_path: str):
        """Builds a list of dirs, each representing an Intent directory.

        Ex: /path/to/agent/intents/<intent_dir>

        This dir path can be used to find the next level of information
        in the directory by appending the appropriate next dir structures like:
        - <intent_name>.json, for the Intent object metadata
        - /trainingPhrases, for the Training Phrases dir
        """
        root_dir = agent_local_path + "/intents"

        intent_paths = []

        for intent_dir in os.listdir(root_dir):
            intent_dir_path = f"{root_dir}/{intent_dir}"
            intent_paths.append(intent_dir_path)

        return intent_paths

    def check_intent_filters(self, intent: Intent) -> Intent:
        """Determines if the Intent should be filtered for linting."""
        if self.include_filter:
            if self.include_filter in intent.display_name:
                intent.filtered = False

            else:
                intent.filtered = True

        if self.exclude_filter:
            if self.exclude_filter in intent.display_name:
                intent.filtered = True

        return intent

    def lint_intent_metadata(self, intent: Intent, stats: LintStats):
        """Lint the metadata file for a single Intent."""
        intent.metadata_file = f"{intent.dir_path}/{intent.display_name}.json"

        try:
            with open(intent.metadata_file, "r", encoding="UTF-8") as meta_file:
                intent.data = json.load(meta_file)
                intent.resource_id = intent.data.get("name", None)
                intent.labels = intent.data.get("labels", None)
                intent.description = intent.data.get("description", None)

                # TODO: Linting rules for Intent Metadata

                meta_file.close()

        except FileNotFoundError:
            stats = self.rules.intent_missing_metadata(intent, stats)

        return stats

    def lint_language_codes(self, intent: Intent, stats: LintStats):
        """Executes all Training Phrase based linter rules."""

        for lang_code in intent.training_phrases:
            tp_file = Common.get_file_based_on_lang_code_filter(
                intent, lang_code, self.lang_code_filter
            )

            if tp_file:
                with open(tp_file, "r", encoding="UTF-8") as tps:
                    data = json.load(tps)
                    phrases = data.get("trainingPhrases", None)
                    intent.training_phrases[lang_code]["tps"] = phrases

                    stats = self.rules.run_training_phrase_rules(
                        intent, lang_code, stats)

                    tps.close()

        return stats

    def lint_training_phrases(self, intent: Intent, stats: LintStats):
        """Lint the Training Phrase dir for a single Intent."""
        if "trainingPhrases" in os.listdir(intent.dir_path):
            self.build_lang_code_paths(intent)
            stats = self.lint_language_codes(intent, stats)

        else:
            stats = self.rules.missing_training_phrases(intent, stats)

        return stats

    def lint_intent(self, intent: Intent, stats: LintStats):
        """Lint a single Intent directory and associated files."""
        intent.display_name = Common.parse_filepath(intent.dir_path, "intent")
        intent = self.check_intent_filters(intent)

        if not intent.filtered:
            stats.total_intents += 1
            stats = self.lint_intent_metadata(intent, stats)
            stats = self.lint_training_phrases(intent, stats)

        return stats

    def lint_intents_directory(self, agent_local_path: str):
        """Linting the top level Intents Dir in the JSON Package structure.

        The following files/dirs exist under the `intents` dir:
        - <intent_display_name> Directory
          - trainingPhrases
            - <language-code>.json
          - <intent_display_name> Object

        In Dialogflow CX, the Training Phrases of each Intent are stored in
        individual .json files by language code under each Intent Display
        Name. In this method, we will lint all Intent dirs, including the
        training phrase files and metadata objects for each Intent.
        """
        start_message = f'{"#" * 10} Begin Intents Directory Linter'
        self.console.log(start_message)

        stats = LintStats()

        # Create a list of all Intent paths to iter through
        intent_paths = self.build_intent_path_list(agent_local_path)

        # Linting Starts Here
        for intent_path in intent_paths:
            intent = Intent()
            intent.verbose = self.verbose
            intent.agent_id = self.agent_id
            intent.dir_path = intent_path

            intent = self.load_naming_conventions(
                intent, self.naming_conventions)
            stats = self.lint_intent(intent, stats)

        header = "-" * 20
        rating = Common.calculate_rating(
            stats.total_issues, stats.total_inspected
        )

        end_message = (
            f"\n{header}\n{stats.total_intents} Intents linted."
            f"\n{stats.total_issues} issues found out of "
            f"{stats.total_inspected} inspected."
            f"\nYour Agent Intents rated at {rating:.2f}/10\n\n"
        )
        self.console.log(end_message)



"""Flow Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any
from cxlint.resources.types import Flow, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class FlowRules:
    """Flow Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

    # naming-conventions
    def flow_naming_convention(
            self,
            flow: Flow,
            stats: LintStats) -> LintStats:
        """Check that the Flow Display Name conforms to naming conventions."""
        rule = "R015: Naming Conventions"

        if flow.naming_pattern:
            res = re.search(flow.naming_pattern, flow.display_name)

            stats.total_inspected += 1

        if not res:
            resource = Resource()
            resource.agent_id = flow.agent_id
            resource.flow_display_name = flow.display_name
            resource.flow_id = flow.resource_id
            resource.page_display_name = None
            resource.page_id = None
            resource.resource_type = "flow"

            message = ": Flow Display Name does not meet the specified Naming"\
                  f" Convention : {flow.naming_pattern}"
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # unused-pages
    def unused_pages(self, flow: Flow, stats: LintStats) -> LintStats:
        """Checks for Unusued Pages in Flow Graph."""
        rule = "R012: Unused Pages"

        for page in flow.unused_pages:
            resource = Resource()
            resource.agent_id = flow.agent_id
            resource.flow_display_name = flow.display_name
            resource.flow_id = flow.resource_id
            resource.page_display_name = page
            resource.page_id = flow.data.get(page, None)
            resource.resource_type = "page"

            message = ""
            stats.total_inspected += 1
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # dangling-pages
    def dangling_pages(self, flow: Flow, stats: LintStats) -> LintStats:
        """Checks for Dangling Pages in Flow Graph."""
        rule = "R013: Dangling Pages"

        for page in flow.dangling_pages:
            resource = Resource()
            resource.agent_id = flow.agent_id
            resource.flow_display_name = flow.display_name
            resource.flow_id = flow.resource_id
            resource.page_display_name = page
            resource.page_id = flow.data.get(page, None)
            resource.resource_type = "page"

            message = ""
            stats.total_inspected += 1
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # unreachable-pages
    def unreachable_pages(self, flow: Flow, stats: LintStats) -> LintStats:
        """Checks for Unreachable Pages in Flow Graph."""
        rule = "R014: Unreachable Pages"

        for page in flow.unreachable_pages:
            resource = Resource()
            resource.agent_id = flow.agent_id
            resource.flow_display_name = flow.display_name
            resource.flow_id = flow.resource_id
            resource.page_display_name = page
            resource.page_id = flow.data.get(page, None)
            resource.resource_type = "page"

            message = ""
            stats.total_inspected += 1
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # extra-display-name-whitespace
    def flow_display_name_extra_whitespaces(
        self,
        flow: Flow,
        stats: LintStats) -> LintStats:
        """Check Flow display name for extra whitespace characters."""
        rule = "R016: Extra Whitespace in Display Name"

        res = bool(flow.display_name.startswith(" ") or
                   flow.display_name.endswith(" ") or
                   re.search(r"\s{2,}", flow.display_name))

        if res :
            resource = Resource()
            resource.agent_id = flow.agent_id
            resource.flow_display_name = flow.display_name
            resource.flow_id = flow.resource_id
            resource.page_display_name = None
            resource.page_id = None
            resource.resource_type = "flow"

            message = ""
            stats.total_issues += 1
            stats.total_inspected += 1

            self.log.generic_logger(resource, rule, message)

        return stats


    def run_flow_rules(self, flow: Flow, stats: LintStats) -> LintStats:
        """Checks and Executes all Flow level rules."""
        # naming-conventions
        if self.disable_map.get("naming-conventions", True):
            stats = self.flow_naming_convention(flow, stats)

        # unused-pages
        if self.disable_map.get("unused-pages", True):
            stats = self.unused_pages(flow, stats)

        # dangling-pages
        if self.disable_map.get("dangling-pages", True):
            stats = self.dangling_pages(flow, stats)

        # unreachable-pages
        if self.disable_map.get("unreachable-pages", True):
            stats = self.unreachable_pages(flow, stats)

        # extra-display-name-whitespace
        if self.disable_map.get("extra-display-name-whitespace", True):
            stats = self.flow_display_name_extra_whitespaces(flow, stats)


        return stats



"""Intent Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any, List, Union
from cxlint.resources.types import Intent, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class IntentRules:
    """Intent Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

    @staticmethod
    def check_if_head_intent(intent: Intent) -> bool:
        """Checks if Intent is Head Intent based on labels and name."""
        hid = False

        if "head" in intent.display_name:
            hid = True

        return hid

    @staticmethod
    def check_if_confirmation_intent(tps: List[str]) -> bool:
        """Check if the Intent contains yes/no phrases for confirmation."""
        confirm = False
        confirm_set = set(["yes", "no"])

        res = confirm_set.intersection(set(tps))
        if res:
            confirm = True

        return confirm

    @staticmethod
    def check_if_escalation_intent(tps: List[str]) -> bool:
        """Check if the Intent contains escalation phrases."""
        escalate = False
        escalate_set = set(["escalate", "operator"])

        res = escalate_set.intersection(set(tps))
        if res:
            escalate = True

        return escalate

    @staticmethod
    def flatten_training_phrase_parts(parts: Dict[str, Any]) -> List[str]:
        """Flatten the Training Phrase Parts proto to list of strings."""

        # If the TP Part has more than 1 part, we need to extract and
        # concat the data into a single string
        if len(parts) > 1:
            utterance = ""
            for part in parts:
                utterance += part.get("text", "")

        # Otherwise, there's just 1 part so we can take it as-is
        else:
            utterance = parts[0].get("text", None)

        return utterance

    def gather_training_phrases(
        self, intent: Intent, lang_code: str) -> List[str]:
        """Flatten the Training Phrase proto to a list of strings."""
        tps_flat = []
        tps_original = intent.training_phrases.get(lang_code, None)["tps"]

        for tp in tps_original:
            parts = tp.get("parts", None)
            utterance = self.flatten_training_phrase_parts(parts)
            tps_flat.append(utterance)

        return tps_flat

    def check_and_log_naming(
        self,
        intent: Intent,
        stats: LintStats,
        res: Union[re.Match, None],
        pattern: str):
        """Checks for final naming match and calls logger."""
        rule = "R015: Naming Conventions"

        if not res:
            resource = Resource()
            resource.agent_id = intent.agent_id
            resource.intent_display_name = intent.display_name
            resource.intent_id = intent.resource_id
            resource.resource_type = "intent"

            message = ": Intent Display Name does not meet the specified"\
                f" Convention : {pattern}"

            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # naming-conventions
    def intent_naming_convention(
        self,
        intent: Intent,
        lang_code: str,
        stats: LintStats) -> LintStats:
        """Check that the Display Name conforms to naming conventions."""

        hid = self.check_if_head_intent(intent)
        tps = self.gather_training_phrases(intent, lang_code)
        confirm = self.check_if_confirmation_intent(tps)
        escalate = self.check_if_escalation_intent(tps)

        # Head Intents
        if hid and intent.naming_pattern_head:
            pattern = intent.naming_pattern_head
            res = re.search(pattern, intent.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(intent, stats, res, pattern)

        # Confirmation Intents
        elif confirm and intent.naming_pattern_confirmation:
            pattern = intent.naming_pattern_confirmation
            res = re.search(pattern, intent.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(intent, stats, res, pattern)

        # Escalation Intents
        elif escalate and intent.naming_pattern_escalation:
            pattern = intent.naming_pattern_escalation
            res = re.search(pattern, intent.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(intent, stats, res, pattern)

        # Generic Intents
        elif intent.naming_pattern_generic:
            pattern = intent.naming_pattern_generic
            res = re.search(pattern, intent.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(intent, stats, res, pattern)

        return stats


    # intent-missing-metadata
    def intent_missing_metadata(
        self,
        intent: Intent,
        stats: LintStats) -> LintStats:
        """Flags Intent that has missing metadata file.

        This rule is separate from the main group of Intent rules because it
        checks for the existence of Intent files and metadata prior to
        unpacking any language codes or training phrases. It will not be
        bundled in the `run_training_phrase_rules` method.
        """
        rule = "R010: Missing Metadata file for Intent"
        message = ""

        resource = Resource()
        resource.agent_id = intent.agent_id
        resource.intent_display_name = intent.display_name
        resource.intent_id = intent.resource_id
        resource.resource_type = "intent"

        stats.total_inspected += 1
        stats.total_issues += 1

        self.log.generic_logger(resource, rule, message)

        return stats

    # intent-missing-tps
    def missing_training_phrases(
        self,
        intent: Intent,
        stats: LintStats) -> LintStats:
        """Checks for Intents that are Missing Training Phrases

        This rule is separate from the main group of Intent rules because it
        checks for the existence of data inside a single language code file.
        If data is completely missing (i.e. no training phrases), the file
        is never opened and we only flag the missing phrases.
        """
        if self.disable_map.get("intent-missing-tps", True):
            rule = "R004: Intent is Missing Training Phrases."
            message = f": {intent.training_phrases}"

            resource = Resource()
            resource.agent_id = intent.agent_id
            resource.intent_display_name = intent.display_name
            resource.intent_id = intent.resource_id
            resource.resource_type = "intent"

            stats.total_inspected += 1
            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        return stats

    # intent-min-tps
    def min_tps_head_intent(
        self,
        intent: Intent,
        lang_code: str,
        stats: LintStats) -> LintStats:
        """Determines if Intent has min recommended training phrases"""
        n_tps = len(intent.training_phrases[lang_code]["tps"])
        stats.total_inspected += 1

        hid = self.check_if_head_intent(intent)

        resource = Resource()
        resource.agent_id = intent.agent_id
        resource.intent_display_name = intent.display_name
        resource.intent_id = intent.resource_id
        resource.resource_type = "intent"

        if hid and n_tps < 50:
            rule = "R005: Head Intent Does Not Have Minimum Training Phrases."
            message = f": {lang_code} : ({n_tps} / 50)"

            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        elif n_tps < 20:
            rule = "R005: Intent Does Not Have Minimum Training Phrases."
            message = f": {lang_code} : ({n_tps} / 20)"

            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        return stats

    # extra-display-name-whitespace
    def intent_display_name_extra_whitespaces(
        self,
        intent: Intent,
        stats: LintStats) -> LintStats:
        """Check Intent display name for extra whitespace characters."""
        rule = "R016: Extra Whitespace in Display Name"

        stats.total_inspected += 1

        res = bool(intent.display_name.startswith(" ") or
                   intent.display_name.endswith(" ") or
                   re.search(r"\s{2,}", intent.display_name))

        if res :
            resource = Resource()
            resource.agent_id = intent.agent_id
            resource.intent_display_name = intent.display_name
            resource.intent_id = intent.resource_id
            resource.resource_type = "intent"

            message = ""
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats


    def run_training_phrase_rules(
        self,
        intent: Intent,
        lang_code: str,
        stats: LintStats) -> LintStats:
        """Checks and Executes all Intent/Training Phrase level rules.

        The requirements for a rule in this section are:
          - Intent must have at least 1 langauge_code file
          - Intent must have at least 1 training phrase
        """
        # naming-conventions
        if self.disable_map.get("naming-conventions", True):
            stats = self.intent_naming_convention(intent, lang_code, stats)

        # intent-min-tps
        if self.disable_map.get("intent-min-tps", True):
            stats = self.min_tps_head_intent(intent, lang_code, stats)

        # extra-display-name-whitespace
        if self.disable_map.get("extra-display-name-whitespace", True):
            stats = self.intent_display_name_extra_whitespaces(intent, stats)

        return stats



"""Agent Metadata linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser

from cxlint.common import Common



class Agents:
    """Agent Metadata linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.config = config
        self.agent_type = Common.load_agent_type(config)
        self.disable_map = Common.load_message_controls(config)
        self.agent_id = Common.load_agent_id(config)
        self.naming_conventions = Common.load_naming_conventions(config)

    def lint_agents_metadata(self, agent_local_path: str):
        """Linting the Agent Metadata file."""

        # TODO (pmarlow):  Add Agent Metadata lint methods



"""Collection of Type Classes used for CX Linter."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Any
from dataclasses import dataclass, field

from cxlint.graph import Graph

@dataclass
class AgentMetadata:
    """Used to track the current Agent Metadata attrinbutes."""

    default_language_code: str = None
    dtmf_settings: bool = False
    logging_enabled: bool = False
    speech_adaptation: bool = False


@dataclass
class Flow:
    """Used to track current Flow Attributes."""

    agent_id: str = None
    all_pages: set = field(default_factory=set)
    active_pages: set = field(default_factory=set)
    data: Dict[str, Any] = field(default_factory=dict)
    dangling_pages: set = field(default_factory=set)
    dir_path: str = None  # Full Directory Path for this Flow
    display_name: str = None  # Flow Display Name (removed special chars)
    file_name: str = None  # Original Name of Flow (includes special chars)
    filtered: bool = False
    graph: Graph = None
    naming_pattern: str = None
    resource_id: str = None
    resource_type: str = "flow"
    start_page_file: str = None  # File Path Location of START_PAGE
    unreachable_pages: set = field(default_factory=set)
    unused_pages: set = field(default_factory=set)
    verbose: bool = False


@dataclass
class Page:
    """Used to track current Page Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None
    entry: Dict[str, Any] = None
    events: List[object] = None
    flow: Flow = None
    form: Dict[str, Any] = None
    has_webhook: bool = False
    has_webhook_event_handler: bool = False
    naming_pattern_generic: str = None
    naming_pattern_form: str = None
    naming_pattern_webhook: str = None
    page_file: str = None
    resource_id: str = None
    resource_type: str = "page"
    routes: List[object] = None
    route_groups: List[str] = None
    verbose: bool = False

@dataclass
class FormParameter:
    """Tracks Form Paramter attributes within a Page."""

    advanced_settings: str = None
    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None
    dtmf_settings: str = None
    entity_type: str = None
    fill_behavior: Dict[str, Any] = None
    init_fulfillment: Dict[str, Any] = None
    page: Page = None
    reprompt_handlers: Dict[str, Any] = None
    required: bool = True


@dataclass
class RouteGroup:
    """Used to track current RouteGroup Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None
    flow: Flow = None
    resource_id: str = None
    resource_type: str = "route_group"
    rg_file: str = None
    routes: List[object] = None
    verbose: bool = False


@dataclass
class Fulfillment:
    """Used to track current Fulfillment Attributes."""

    agent_id: str = None
    agent_type: str = None
    data: Dict[str, Any] = None
    display_name: str = None  # Inherit from Page easy logging
    fulfillment_type: str = None  # transition_route | event
    page: Page = None
    parameter: str = None # Used for Reprompt Event Handlers
    target_flow: str = None
    target_page: str = None
    text: str = None
    trigger: str = None
    resource_type: str = "fulfillment"
    verbose: bool = False


@dataclass
class Intent:
    """Used to track current Intent Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    description: str = None
    display_name: str = None
    dir_path: str = None
    filtered: bool = False
    labels: Dict[str, str] = None
    metadata_file: str = None
    naming_pattern_generic: str = None
    naming_pattern_head: str = None
    naming_pattern_confirmation: str = None
    naming_pattern_escalation: str = None
    resource_id: str = None
    resource_type: str = "intent"
    training_phrases: Dict[str, Any] = field(default_factory=dict)
    verbose: bool = False


@dataclass
class EntityType:
    """Used to track current Flow Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    dir_path: str = None  # Full Directory Path for this Entity Type
    display_name: str = None  # Entity Type Display Name
    entities: Dict[str, Any] = field(default_factory=dict)  # Map
    kind: str = None  # The kind of Entity Type represented
    naming_pattern: str = None
    resource_id: str = None
    resource_type: str = "entity_type"
    verbose: bool = False


@dataclass
class TestCase:
    """Used to track current Test Case Attributes."""

    associated_intent_data: Dict[str, Any] = None
    agent_id: str = None
    agent_path: str = None
    conversation_turns: List[Any] = None
    data: Dict[str, Any] = None
    dir_path: str = None
    display_name: str = None
    has_invalid_intent: bool = False
    intent_data: List[str] = None
    naming_pattern: str = None
    qualified: bool = False
    resource_id: str = None
    resource_type: str = "test_case"
    tags: List[str] = None
    test_config: Dict[str, Any] = None
    verbose: bool = False

@dataclass
class Webhook:
    """Used to track current Webhook attributes."""

    agent_id: str = None
    agent_path: str = None
    data: Dict[str, Any] = None
    dir_path: str = None
    display_name: str = None
    naming_pattern: str = None
    resource_id: str = None
    resource_type: str = "webhook"
    service_type: str = None
    timeout: int = 0
    verbose: bool = False

@dataclass
class Resource:
    """Generic class to store basic Resource data.

    Since each core class has such varied parameters, this generic class will
    help to standardize the data fed to the generic logger and the maps used
    therein.
    """

    agent_id: str = None
    entity_type_display_name: str = None
    entity_type_id: str = None
    flow_display_name: str = None
    flow_id: str = None
    intent_display_name: str = None
    intent_id: str = None
    page_display_name: str = None
    page_id: str = None
    resource_type: str = None
    test_case_display_name: str = None
    test_case_id: str = None
    webhook_display_name: str = None
    webhook_id: str = None


@dataclass
class LintStats:
    """Used to track linter stats for each section processed."""

    total_issues: int = 0
    total_inspected: int = 0
    total_flows: int = 0
    total_pages: int = 0
    total_intents: int = 0
    total_training_phrases: int = 0
    total_entity_types: int = 0
    total_route_groups: int = 0
    total_test_cases: int = 0
    total_webhooks: int = 0



"""Response Message Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any
from cxlint.resources.types import Fulfillment, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class ResponseMessageRules:
    """Response Message Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

   # RESPONSE MESSAGE RULES
    # closed-choice-alternative
    def closed_choice_alternative_parser(
        self, route: Fulfillment, stats: LintStats
    ) -> LintStats:
        """Identifies a Closed Choice Alternative Question."""
        rule = (
            "R001: Closed-Choice Alternative Missing Intermediate `?` "
            "(A? or B.)"
        )
        message = f": {route.trigger}"

        pattern = r"^(What|Where|When|Who|Why|How|Would) (.*) or (.*)\?$"

        match = re.search(pattern, route.text, flags=re.IGNORECASE)

        if match:
            resource = Resource()
            resource.agent_id = route.agent_id
            resource.flow_display_name = route.page.flow.display_name
            resource.flow_id = route.page.flow.resource_id
            resource.page_display_name = route.page.display_name
            resource.page_id = route.page.resource_id
            resource.resource_type = "fulfillment"

            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        return stats

    # wh-questions
    def wh_questions(self, route: Fulfillment, stats: LintStats) -> LintStats:
        """Identifies a Wh- Question and checks for appropriate punctuation."""
        rule = "R002: Wh- Question Should Use `.` Instead of `?` Punctuation"
        message = f": {route.trigger}"

        pattern = r"^(what|when|where|who|why|how)\b.*\?$"

        match = re.search(pattern, route.text, flags=re.IGNORECASE)

        if match and "event" not in route.trigger:
            resource = Resource()
            resource.agent_id = route.agent_id
            resource.flow_display_name = route.page.flow.display_name
            resource.flow_id = route.page.flow.resource_id
            resource.page_display_name = route.page.display_name
            resource.page_id = route.page.resource_id
            resource.resource_type = "fulfillment"

            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        return stats

    # clarifying-questions
    def clarifying_questions(
        self, route: Fulfillment, stats: LintStats
    ) -> LintStats:
        """Identifies Clarifying Questions that are missing `?` Punctuation."""
        rule = "R003: Clarifying Question Should Use `?` Punctuation"
        message = f": {route.trigger}"

        # updated pattern
        pattern = r"^(what|when|where|who|why|how)\b.*\.$"

        match = re.search(pattern, route.text, flags=re.IGNORECASE)

        if match and "event" in route.trigger:
            resource = Resource()
            resource.agent_id = route.agent_id
            resource.flow_display_name = route.page.flow.display_name
            resource.flow_id = route.page.flow.resource_id
            resource.page_display_name = route.page.display_name
            resource.page_id = route.page.resource_id
            resource.resource_type = "fulfillment"

            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        return stats

    def run_rm_text_rules(
        self,
        route: Fulfillment,
        stats: LintStats) -> LintStats:
        """Checks and Executes all Response Message level rules.

        This set of rules will be executed against the Response Messages that
        are of type `text`. This is equivalent to the "Agent Says" sections of
        the agent design-time console.
        """
        voice = False

        if route.agent_type == "voice":
            voice = True

        # Some rules are only appropriate to lint for Voice agents.
        # For example, rules that deal with SSML, DTMF, STT intonation, etc.
        # We will use the `voice` bool value to filter these rules out for
        # non-voice agents.

        # closed-choice-alternative
        if self.disable_map.get("closed-choice-alternative", True) and voice:
            stats = self.closed_choice_alternative_parser(route, stats)

        # wh-questions
        if self.disable_map.get("wh-questions", True) and voice:
            stats = self.wh_questions(route, stats)

        # clarifying-questions
        if self.disable_map.get("clarifying-questions", True) and voice:
            stats = self.clarifying_questions(route, stats)

        return stats



"""Core class and methods for CX Linter."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import os
import logging

from rich.console import Console
from rich.logging import RichHandler

from typing import List, Any, Union, Dict

from cxlint.common import Common
from cxlint.gcs_utils import GcsUtils

from cxlint.resources.agents import Agents
from cxlint.resources.flows import Flows
from cxlint.resources.entity_types import EntityTypes
from cxlint.resources.intents import Intents
from cxlint.resources.test_cases import TestCases
from cxlint.resources.webhooks import Webhooks

console = Console(
    record=True,
    log_time=False,
    log_path=False,
    width=200,
    color_system="truecolor",
)

keywords = [
    "Flows Directory",
    "Entity Types Directory",
    "Test Cases Directory",
    "Intents Directory",
]
handler = RichHandler(
    enable_link_path=False,
    keywords=keywords,
    show_time=False,
    show_level=False,
    show_path=False,
    tracebacks_word_wrap=False,
)

# logging config
logging.basicConfig(
    level=logging.INFO, format="%(message)s", handlers=[handler], force=True
)

# configparser
config = configparser.ConfigParser()
config.sections()

config_filepath = os.path.join(os.path.dirname(__file__), ".cxlintrc")
with open(config_filepath, encoding="UTF-8") as config_filepath_infile:
    config.read_file(config_filepath_infile)


class CxLint:
    """Core CX Linter methods and functions."""

    def __init__(
        self,
        agent_id: str = None,
        agent_type: str = None,
        intent_include_pattern: str = None,
        intent_exclude_pattern: str = None,
        flow_include_list: List[str] = None,
        flow_exclude_list: List[str] = None,
        language_code: Union[List[str], str] = None,
        load_gcs: bool = False,
        naming_conventions: Dict[str, str] = None,
        output_file: str = None,
        resource_filter: Union[List[str], str] = None,
        test_case_pattern: str = None,
        test_case_tags: Union[List[str], str] = None,
        verbose: bool = True,
    ):
        if load_gcs:
            self.gcs = GcsUtils()

        if agent_id:
            self.update_config("AGENT ID", agent_id)

        if agent_type:
            self.update_config("AGENT TYPE", agent_type)

        if flow_include_list or flow_exclude_list:
            self.update_flows_config(flow_include_list, flow_exclude_list)

        if intent_exclude_pattern or intent_include_pattern:
            self.update_intent_config(
                intent_include_pattern, intent_exclude_pattern
            )

        if language_code:
            self.update_config("INTENTS", language_code)

        if naming_conventions:
            self.update_naming_conventions_config(
                "NAMING CONVENTIONS", naming_conventions)

        if resource_filter:
            self.update_config("AGENT RESOURCES", resource_filter)

        if test_case_pattern:
            self.update_config(
                "TEST CASE DISPLAY NAME PATTERN", test_case_pattern
            )

        if test_case_tags:
            self.update_config("TEST CASE TAGS", test_case_tags)

        self.resource_filter = Common.load_resource_filter(config)
        self.output_file = output_file

        self.agents = Agents(verbose, config, console)
        self.entity_types = EntityTypes(verbose, config, console)
        self.intents = Intents(verbose, config, console)
        self.flows = Flows(verbose, config, console)
        self.test_cases = TestCases(verbose, config, console)
        self.webhooks = Webhooks(verbose, config, console)

    @staticmethod
    def read_and_append_to_config(section: str, key: str, data: Any):
        """Reads the existing config file and appends any new data."""
        existing_data = config[section][key]

        # Check for empty string from file and set to None
        if existing_data != "":
            data = existing_data + "," + data

        config.set(section, key, data)

    @staticmethod
    def transform_list_to_str(data: Union[List[str], str]):
        """Determine input data and parse accordingly for config update."""
        res = data

        if isinstance(data, List):
            res = ",".join(data)

        if not isinstance(res, str):
            raise TypeError(
                "Input must be one of the following formats: `str` | "
                "List[`str`]"
            )

        return res

    def update_naming_conventions_config(
        self, section: str, styles: Dict[str, Dict]):
        """Update the Naming Conventions config based on user inputs."""

        for key, value in styles.items():
            if not isinstance(value, str):
                raise TypeError(
                    "Naming Convention values must be type `string`")

            config.set(section, key, value)


    def update_flows_config(self, include_pattern: str, exclude_pattern: str):
        """Handle updates to the Flow include/exclude lists."""
        if include_pattern:
            data = self.transform_list_to_str(include_pattern)
            config.set("FLOWS", "include", data)

        if exclude_pattern:
            data = self.transform_list_to_str(exclude_pattern)
            config.set("FLOWS", "exclude", data)

    def update_intent_config(self, include_pattern: str, exclude_pattern: str):
        """Handle updates to the Intent include/exclude lists."""
        if include_pattern:
            config.set("INTENTS", "include", include_pattern)

        if exclude_pattern:
            config.set("INTENTS", "exclude", exclude_pattern)

    def update_config(self, section: str, data: Any):
        """Update the Config file based on user provided kwargs."""
        if section == "AGENT ID":
            config.set(section, "id", data)

        if section == "AGENT RESOURCES":
            data = self.transform_list_to_str(data)
            config.set(section, "include", data)

        if section == "AGENT TYPE":
            config.set(section, "type", data)

        if section == "INTENTS":
            data = self.transform_list_to_str(data)
            config.set(section, "language_code", data)

        if section == "TEST CASE TAGS":
            data = self.transform_list_to_str(data)
            self.read_and_append_to_config(section, "include", data)

        if section == "TEST CASE DISPLAY NAME PATTERN":
            config.set(section, "pattern", data)

    def lint_agent(self, agent_local_path: str):
        """Linting the entire CX Agent and all resource directories."""
        # agent_file = agent_local_path + '/agent.json'
        # with open(agent_file, 'r', encoding='UTF-8') as agent_data:
        #     data = json.load(agent_data)

        start_message = f'{"=" * 5} LINTING AGENT {"=" * 5}\n'
        console.log(start_message)
        resources = Common.resource_precheck(
            agent_local_path, self.resource_filter)

        if resources["agents"]:
            self.agents.lint_agents_metadata(agent_local_path)

        if resources["flows"]:
            self.flows.lint_flows_directory(agent_local_path)

        if resources["entity_types"]:
            self.entity_types.lint_entity_types_directory(agent_local_path)

        if resources["intents"]:
            self.intents.lint_intents_directory(agent_local_path)

        if resources["test_cases"]:
            self.test_cases.lint_test_cases_directory(agent_local_path)

        if resources["webhooks"]:
            self.webhooks.lint_webhooks_directory(agent_local_path)

        if self.output_file:
            console.save_text(self.output_file)



"""Page Level Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any, Union
from cxlint.resources.types import Page, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class PageRules:
    """Page Level Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

    @staticmethod
    def _gather_params_and_handlers(parameter):
        """Check to see if Reprompt Event Handlers exist."""
        fill = parameter.get("fillBehavior", None)
        if fill:
            handlers = fill.get("repromptEventHandlers", [])

        param_handler = {
            "display_name": parameter.get("displayName", None),
            "handlers": handlers
        }

        return param_handler

    def check_and_log_naming(
        self,
        page: Page,
        stats: LintStats,
        res: Union[re.Match, None],
        pattern: str):
        """Checks for final naming match and calls logger."""
        rule = "R015: Naming Conventions"

        if not res:
            resource = Resource()
            resource.agent_id = page.agent_id
            resource.flow_display_name = page.flow.display_name
            resource.flow_id = page.flow.resource_id
            resource.page_display_name = page.display_name
            resource.page_id = page.resource_id
            resource.resource_type = "page"

            message = ": Page Display Name does not meet the specified"\
                f" Convention : {pattern}"

            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # naming-conventions
    def page_naming_conventions(
        self, page: Page, stats: LintStats) -> LintStats:
        """Check that the Page Display Name conform to naming conventions."""

        # Return early if Start Page
        if page.display_name == "Start Page":
            return stats

        # Form Pages
        if page.form and page.naming_pattern_form:
            pattern = page.naming_pattern_form
            res = re.search(pattern, page.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(page, stats, res, pattern)

        # Webhook Pages
        elif page.has_webhook and page.naming_pattern_webhook:
            pattern = page.naming_pattern_webhook
            res = re.search(pattern, page.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(page, stats, res, pattern)

        # Generic Pages
        elif page.naming_pattern_generic:
            pattern = page.naming_pattern_generic
            res = re.search(pattern, page.display_name)
            stats.total_inspected += 1

            stats = self.check_and_log_naming(page, stats, res, pattern)

        return stats

    # missing-webhook-event-handlers
    def missing_webhook_event_handlers(
        self, page: Page, stats: LintStats) -> LintStats:
        """Checks for missing Event Handlers on pages that use Webhooks."""
        rule = "R011: Missing Webhook Event Handlers"

        stats.total_inspected += 1

        if page.has_webhook and not page.has_webhook_event_handler:
            resource = Resource()
            resource.agent_id = page.agent_id
            resource.flow_display_name = page.flow.display_name
            resource.flow_id = page.flow.resource_id
            resource.page_display_name = page.display_name
            resource.page_id = page.resource_id
            resource.resource_type = "page"

            message = ""

            stats.total_issues += 1
            self.log.generic_logger(resource, rule, message)

        return stats

    # extra-display-name-whitespace
    def page_display_name_extra_whitespaces(
        self,
        page: Page,
        stats: LintStats) -> LintStats:
        """Check Page display name for extra whitespace characters."""
        rule = "R016: Extra Whitespace in Display Name"

        stats.total_inspected += 1

        res = bool(page.display_name.startswith(" ") or
                   page.display_name.endswith(" ") or
                   re.search(r"\s{2,}", page.display_name))

        if res :
            resource = Resource()
            resource.agent_id = page.agent_id
            resource.flow_display_name = page.flow.display_name
            resource.flow_id = page.flow.resource_id
            resource.page_display_name = page.display_name
            resource.page_id = page.resource_id
            resource.resource_type = "page"

            message = ""
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # page-form-no-match-handler
    def page_form_no_match_handler(
        self, page: Page, stats: LintStats) -> LintStats:
        """Check Page Form Parameters for NO_MATCH handlers."""
        rule = "R017: Missing NO_MATCH Handlers on Form"

        stats.total_inspected += 1

        params = []

        if page.form:
            for parameter in page.form["parameters"]:
                params.append(self._gather_params_and_handlers(parameter))

        for param in params:
            if not param.get("handlers", None):
                resource = Resource()
                resource.agent_id = page.agent_id
                resource.flow_display_name = page.flow.display_name
                resource.flow_id = page.flow.resource_id
                resource.page_display_name = page.display_name
                resource.page_id = page.resource_id
                resource.resource_type = "page"

                message = f': {param.get("display_name", None)}'
                stats.total_issues += 1

                self.log.generic_logger(resource, rule, message)

        return stats

    # page-form-no-input-handler
    def page_form_no_input_handler(
        self, page: Page, stats: LintStats) -> LintStats:
        """Check Page Form Parameters for NO_INPUT handlers."""
        rule = "R018: Missing NO_INPUT Handlers on Form"

        stats.total_inspected += 1

        params = []

        if page.form:
            for parameter in page.form["parameters"]:
                params.append(self._gather_params_and_handlers(parameter))

        for param in params:
            if not param.get("handlers", None):
                resource = Resource()
                resource.agent_id = page.agent_id
                resource.flow_display_name = page.flow.display_name
                resource.flow_id = page.flow.resource_id
                resource.page_display_name = page.display_name
                resource.page_id = page.resource_id
                resource.resource_type = "page"

                message = f': {param.get("display_name", None)}'
                stats.total_issues += 1

                self.log.generic_logger(resource, rule, message)

        return stats

    def run_page_rules(self, page: Page, stats: LintStats):
        """Checks and Executes all Page level rules."""
        # naming-conventions
        if self.disable_map.get("naming-conventions", True):
            stats = self.page_naming_conventions(page, stats)

        # missing-webhook-event-handlers
        if self.disable_map.get("missing-webhook-event-handlers", True):
            stats = self.missing_webhook_event_handlers(page, stats)

        # extra-display-name-whitespace
        if self.disable_map.get("extra-display-name-whitespace", True):
            stats = self.page_display_name_extra_whitespaces(page, stats)

        # page-form-no-match-handler
        if self.disable_map.get("page-form-no-match-handler", True):
            self.page_form_no_match_handler(page,stats)

        # page-form-no-input-handler
        if self.disable_map.get("page-form-no-input-handler", True):
            self.page_form_no_input_handler(page,stats)

        return stats



"""Test Case Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any
from cxlint.resources.types import TestCase, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class TestCaseRules:
    """Test Case Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

    # naming-conventions
    def test_case_naming_convention(
        self, tc:TestCase, stats: LintStats) -> LintStats:
        """Check Test Case Display Name conforms to naming conventions."""
        rule = "R015: Naming Conventions"

        if tc.naming_pattern:
            res = re.search(tc.naming_pattern, tc.display_name)

            stats.total_inspected += 1

        if not res:
            resource = Resource()
            resource.agent_id = tc.agent_id
            resource.test_case_display_name = tc.display_name
            resource.test_case_id = tc.resource_id
            resource.resource_type = "test_case"

            message = ": Test Case Display Name does not meet the specified"\
                f" Convention : {tc.naming_pattern}"
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # explicit-tps-in-test-cases
    def explicit_tps_in_tcs(self, tc: TestCase, stats: LintStats) -> LintStats:
        """Checks that user utterance is an explicit intent training phrase."""
        rule = "R007: Explicit Training Phrase Not in Test Case"

        for pair in tc.intent_data:
            stats.total_inspected += 1

            intent = pair["intent"]
            phrase = pair["user_utterance"]

            if phrase not in pair["training_phrases"]:
                message = f": [Utterance: {phrase} | Intent: {intent}]"

                resource = Resource()
                resource.agent_id = tc.agent_id
                resource.test_case_display_name = tc.display_name
                resource.test_case_id = tc.resource_id
                resource.resource_type = "test_case"

                stats.total_issues += 1
                self.log.generic_logger(resource, rule, message)

        return stats

    # invalid-intent-in-test-cases
    def invalid_intent_in_tcs(
        self, tc: TestCase, stats: LintStats
    ) -> LintStats:
        """Check that a listed Intent in the Test Case exists in the agent."""
        rule = "R008: Invalid Intent in Test Case"

        for pair in tc.intent_data:
            if pair["status"] == "invalid_intent":
                stats.total_inspected += 1
                stats.total_issues += 1

        resource = Resource()
        resource.agent_id = tc.agent_id
        resource.test_case_display_name = tc.display_name
        resource.test_case_id = tc.resource_id
        resource.resource_type = "test_case"

        message = ""
        self.log.generic_logger(resource, rule, message)

        return stats

    def run_test_case_rules(self, tc: TestCase, stats: LintStats) -> LintStats:
        """Checks and Executes all Test Case level rules."""
        # naming-conventions
        if self.disable_map.get("naming-conventions", True):
            stats = self.test_case_naming_convention(tc, stats)

        # explicit-tps-in-test-cases
        if tc.qualified:
            if self.disable_map.get("explicit-tps-in-test-cases", True):
                stats.total_test_cases += 1
                stats = self.explicit_tps_in_tcs(tc, stats)

        # invalid-intent-in-test-cases
        if tc.has_invalid_intent:
            if self.disable_map.get("invalid-intent-in-test-cases", True):
                stats.total_test_cases += 1
                stats = self.invalid_intent_in_tcs(tc, stats)

        return stats



"""Webhook Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any
from cxlint.resources.types import Webhook, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class WebhookRules:
    """Webhook Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

    # naming-conventions
    def webhook_naming_conventions(
        self, webhook: Webhook, stats: LintStats) -> LintStats:
        """Check the Webhook Display Name conforms to naming convention."""
        rule = "R015: Naming Conventions"

        if webhook.naming_pattern:
            res = re.search(webhook.naming_pattern, webhook.display_name)

            stats.total_inspected += 1

        if not res:
            resource = Resource()
            resource.agent_id = webhook.agent_id
            resource.webhook_display_name = webhook.display_name
            resource.webhook_id = webhook.resource_id
            resource.resource_type = "webhook"

            message = ": Webhook Display Name does not meet the specified"\
                f" Naming Convention : {webhook.naming_pattern}"
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    def run_webhook_rules(
        self, webhook: Webhook, stats: LintStats) -> LintStats:
        """Checks and Executes all Webhook level rules."""

        # naming-conventions
        if self.disable_map.get("naming-conventions", True):
            stats = self.webhook_naming_conventions(webhook, stats)

        return stats



"""Test Case linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from configparser import ConfigParser
from typing import Dict, List, Any

from cxlint.common import Common
from cxlint.rules.test_cases import TestCaseRules
from cxlint.resources.types import TestCase, LintStats


class TestCases:
    """Test Case linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.agent_id = Common.load_agent_id(config)
        self.disable_map = Common.load_message_controls(config)
        self.naming_conventions = Common.load_naming_conventions(config)
        self.tag_filter = self.load_tag_filter(config)
        self.display_name_filter = self.load_display_name_filter(config)

        self.rules = TestCaseRules(console, self.disable_map)

    @staticmethod
    def load_tag_filter(config: ConfigParser) -> Dict[str, str]:
        """Loads the config file for test cases into a map."""
        tag_list = (
            config["TEST CASE TAGS"]["include"].replace("\n", "").split(",")
        )

        # Check for empty tag_list from file and set to None
        if len(tag_list) == 1 and tag_list[0] == "":
            tag_list = None

        else:
            for i, tag in enumerate(tag_list):
                if tag[0] != "#":
                    tag = f"#{tag}"
                    tag_list[i] = tag

        return tag_list

    @staticmethod
    def load_display_name_filter(config: ConfigParser) -> str:
        """Loads the matching pattern for test case display names."""
        pattern = config["TEST CASE DISPLAY NAME PATTERN"]["pattern"]

        return pattern

    @staticmethod
    def build_test_case_path_list(agent_local_path: str):
        """Builds a list of files, each representing a test case."""
        root_dir = agent_local_path + "/testCases"

        test_case_paths = []

        for test_case in os.listdir(root_dir):
            end = test_case.split(".")[-1]
            if end == "json":
                test_case_path = f"{root_dir}/{test_case}"
                test_case_paths.append(test_case_path)

        return test_case_paths

    @staticmethod
    def get_test_case_intent_phrase_pair(tc: TestCase) -> List[Dict[str, str]]:
        """Parse Test Case and return a list of intents in use.

        This method will produce a List of Dicts where the contents of each
        dict is the Training Phrase and associated Triggered Intent as listed
        in the Test Case Conversation Turn. This information is used to compare
        the User Input training phrase with the actual training phrases that
        exist in the Intent resource.

        The dict format is as follows:
            {
                training_phrase: <training_phrase>,
                intent: <intent_display_name>
            }
        """
        intent_data = []

        if tc.conversation_turns:
            for turn in tc.conversation_turns:
                user = turn["userInput"]
                agent = turn["virtualAgentOutput"]
                intent = agent.get("triggeredIntent", None)
                phrase = user.get("input", None)

                text = phrase.get("text", None)
                # TODO pmarlow: Add DTMF user inputs

                if text:
                    text = text["text"]

                if intent and text:
                    intent_data.append(
                        {
                            "user_utterance": text,
                            "intent": intent["name"],
                            "status": "valid",
                            "training_phrases": [],
                        }
                    )

        return intent_data

    @staticmethod
    def get_test_case_intent_data(agent_local_path: str):
        """Collect all Intent Files and Training Phrases for Test Case."""
        # TODO (pmarlow) consolidate into build_intent_paths

        intents_path = agent_local_path + "/intents"

        intent_paths = []

        for intent_dir in os.listdir(intents_path):
            intent_dir_path = f"{intents_path}/{intent_dir}"
            intent_paths.append(
                {"intent": intent_dir, "file_path": intent_dir_path}
            )

        return intent_paths

    @staticmethod
    def flatten_tp_data(tp_data: List[Any]):
        """Flatten the Training Phrase proto to a list of strings."""
        cleaned_tps = []

        for tp in tp_data["trainingPhrases"]:
            parts_list = [part["text"].lower() for part in tp["parts"]]
            cleaned_tps.append("".join(parts_list))

        return cleaned_tps

    def gather_intent_tps(self, tc: TestCase):
        # TODO Refactor
        """Collect all TPs associated with Intent data in Test Case."""
        tc.associated_intent_data = {}

        for i, pair in enumerate(tc.intent_data):
            intent_dir = tc.agent_path + "/intents/" + pair["intent"]

            try:
                if "trainingPhrases" in os.listdir(intent_dir):
                    training_phrases_path = intent_dir + "/trainingPhrases"

                    for lang_file in os.listdir(training_phrases_path):
                        # lang_code = lang_file.split(".")[0]
                        lang_code_path = f"{training_phrases_path}/{lang_file}"

                        with open(
                            lang_code_path, "r", encoding="UTF-8"
                        ) as tp_file:
                            tp_data = json.load(tp_file)
                            cleaned_tps = self.flatten_tp_data(tp_data)

                            tp_file.close()

                        # TODO pmarlow: refactor to use tc.intent_data instead
                        # Need to create another level inside the Intent Dict
                        # that contains the language files as well.
                        tc.intent_data[i]["training_phrases"].extend(
                            cleaned_tps
                        )
                        tc.associated_intent_data[pair["intent"]] = cleaned_tps

            except FileNotFoundError:
                tc.intent_data[i]["status"] = "invalid_intent"
                tc.has_invalid_intent = True
                continue

        return tc

    def qualify_test_case(self, tc: TestCase):
        """Ensure Test Case meets all required filters and prerequisites."""
        tag_match = True
        display_name_match = False

        # Check to see if TC tags match the provided filter tags. If we fail a
        # match we can return early. If no tag_filter is provided we can skip.
        if tc.tags and self.tag_filter:
            tag_match = bool(set(tc.tags).intersection(set(self.tag_filter)))

        # Check to see if TC display name matches provided display name filter.
        # If the filter is not in the display name, we can return early. If no
        # display_name_filter is provided, we can skip.
        if self.display_name_filter in tc.display_name:
            display_name_match = True

        if all([tag_match, display_name_match]):
            tc.intent_data = self.get_test_case_intent_phrase_pair(tc)
            tc = self.gather_intent_tps(tc)

            if not tc.has_invalid_intent:
                tc.qualified = True

        return tc

    def lint_test_case(self, tc: TestCase, stats: LintStats):
        """Lint a single Test Case file."""

        with open(tc.dir_path, "r", encoding="UTF-8") as tc_file:
            tc.data = json.load(tc_file)
            tc.resource_id = tc.data.get("name", None)
            tc.display_name = tc.data.get("displayName", None)
            tc.tags = tc.data.get("tags", None)
            tc.conversation_turns = tc.data.get(
                "testCaseConversationTurns", None
            )
            tc.test_config = tc.data.get("testConfig", None)

            tc = self.qualify_test_case(tc)

            tc_file.close()

        stats = self.rules.run_test_case_rules(tc, stats)

        return stats

    def lint_test_cases_directory(self, agent_local_path: str):
        """Linting the test cases dir in the JSON package structure."""
        start_message = f'{"#" * 10} Begin Test Cases Directory Linter'
        self.console.log(start_message)

        stats = LintStats()

        test_case_paths = self.build_test_case_path_list(agent_local_path)

        # Linting Starts Here
        for test_case in test_case_paths:
            tc = TestCase()
            tc.verbose = self.verbose
            tc.agent_id = self.agent_id
            tc.dir_path = test_case
            tc.agent_path = agent_local_path
            tc.naming_pattern = self.naming_conventions.get(
                "test_case_name", None)
            stats = self.lint_test_case(tc, stats)

        header = "-" * 20
        rating = Common.calculate_rating(
            stats.total_issues, stats.total_inspected
        )

        end_message = (
            f"\n{header}\n{stats.total_test_cases} Test Cases linted."
            f"\n{stats.total_issues} issues found out of "
            f"{stats.total_inspected} inspected."
            f"\nYour Agent Test Cases rated at {rating:.2f}/10\n\n"
        )
        self.console.log(end_message)



"""Entity Type Rules and Definitions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from typing import Dict, Any, List, Union
from cxlint.resources.types import EntityType, LintStats, Resource

from cxlint.rules.logger import RulesLogger

class EntityTypeRules:
    """Entity Type Rules and Definitions."""
    def __init__(
            self,
            console,
            disable_map: Dict[str, Any]):

        self.console = console
        self.disable_map = disable_map
        self.log = RulesLogger(console=console)

    @staticmethod
    def entity_regex_matching(data: Union[List[str], str]) -> bool:
        """Checks Entities and synonyms for issues based on regex pattern."""
        issue_found = False
        pattern = r"^(?:yes|no)$"

        # Checks the individual Entity key
        if isinstance(data, str):
            data_match = re.search(pattern, data, flags=re.IGNORECASE)
            if data_match:
                issue_found = True

        # Checks the list of synonyms if available
        elif isinstance(data, List):
            n = len(data)
            i = 0
            while i != n:
                data_match = re.search(pattern, data[i], flags=re.IGNORECASE)
                if data_match:
                    issue_found = True
                    break

                i += 1

        return issue_found

    def _yes_no_entity_check(
            self,
            etype: EntityType,
            entity: str,
            lang_code: str,
            stats: LintStats) -> LintStats:
        """Check the Entity inside the Entity Type for yes/no phrases."""
        stats.total_inspected += 1

        issue_found = self.entity_regex_matching(entity)

        if issue_found:
            stats.total_issues += 1
            rule = "R009: Yes/No Entities Present in Agent"
            message = f": {lang_code} : Entity : {entity}"

            resource = Resource()
            resource.agent_id = etype.agent_id
            resource.entity_type_display_name = etype.display_name
            resource.entity_type_id = etype.resource_id
            resource.resource_type = "entity_type"

            self.log.generic_logger(resource, rule, message)

        return stats

    def _yes_no_synonym_check(
        self,
        etype: EntityType,
        synonyms: List[str],
        lang_code: str,
        stats: LintStats) -> LintStats:
        """Check the Synonyms of the Entity for yes/no phrases."""
        stats.total_inspected += 1

        issue_found = self.entity_regex_matching(synonyms)

        if issue_found:
            stats.total_issues += 1
            rule = "R009: Yes/No Entities Present in Agent"
            message = f": {lang_code} : Synonyms : {synonyms}"

            resource = Resource()
            resource.agent_id = etype.agent_id
            resource.entity_type_display_name = etype.display_name
            resource.entity_type_id = etype.resource_id
            resource.resource_type = "entity_type"

            self.log.generic_logger(resource, rule, message)

        return stats

    # naming-conventions
    def entity_type_naming_convention(
        self,
        etype: EntityType,
        stats: LintStats) -> LintStats:
        """Check that the Entity Type display name conform to given pattern."""
        rule = "R015: Naming Conventions"

        if etype.naming_pattern:
            res = re.search(etype.naming_pattern, etype.display_name)

            stats.total_inspected += 1

        if not res:
            resource = Resource()
            resource.agent_id = etype.agent_id
            resource.entity_type_display_name = etype.display_name
            resource.entity_type_id = etype.resource_id
            resource.resource_type = "entity_type"

            message = ": Entity Type Display Name does not meet the specified"\
                f" Convention : {etype.naming_pattern}"
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # extra-display-name-whitespace
    def entity_display_name_extra_whitespaces(
        self,
        etype: EntityType,
        stats: LintStats) -> LintStats:
        """Check Entity display name for extra whitespace characters."""
        rule = "R016: Extra Whitespace in Display Name"

        stats.total_inspected += 1

        res = bool(etype.display_name.startswith(" ") or
                   etype.display_name.endswith(" ") or
                   re.search(r"\s{2,}", etype.display_name))

        if res :
            resource = Resource()
            resource.agent_id = etype.agent_id
            resource.entity_type_display_name = etype.display_name
            resource.entity_type_id = etype.resource_id
            resource.resource_type = "entity_type"

            message = ""
            stats.total_issues += 1

            self.log.generic_logger(resource, rule, message)

        return stats

    # yes-no-entities
    def yes_no_entities(
        self,
        etype: EntityType,
        lang_code: str,
        stats: LintStats) -> LintStats:
        """Check that yes/no Entities or Synonyms aren't used in the agent."""
        for entity in etype.entities[lang_code]["entities"]:
            value = entity["value"]
            synonyms = entity["synonyms"]

            stats = self._yes_no_entity_check(etype, value, lang_code, stats)
            stats = self._yes_no_synonym_check(
                etype, synonyms, lang_code, stats
            )

        return stats

    def run_entity_type_rules(
        self,
        etype: EntityType,
        lang_code: str,
        stats: LintStats) -> LintStats:
        """Checks and Executes all Entity Type level rules."""
        # naming-conventions
        if self.disable_map.get("naming-conventions", True):
            stats = self.entity_type_naming_convention(etype, stats)

        # yes-no-entities
        if self.disable_map.get("yes-no-entities", True):
            stats = self.yes_no_entities(etype, lang_code, stats)

        # extra-display-name-whitespace
        if self.disable_map.get("extra-display-name-whitespace", True):
            stats = self.entity_display_name_extra_whitespaces(etype, stats)

        return stats



"""Entity Type linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from configparser import ConfigParser

from cxlint.common import Common
from cxlint.rules.entity_types import EntityTypeRules
from cxlint.resources.types import EntityType, LintStats


class EntityTypes:
    """Entity Type linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.config = config
        self.agent_id = Common.load_agent_id(config)
        self.disable_map = Common.load_message_controls(config)
        self.lang_code_filter = Common.load_lang_code_filter(config)
        self.naming_conventions = Common.load_naming_conventions(config)
        self.rules = EntityTypeRules(console, self.disable_map)

    @staticmethod
    def build_entity_type_path_list(agent_local_path: str):
        """Builds a list of dirs, each representing an Entity Type directory.

        Ex: /path/to/agent/entityTypes/<entity_type_dir>

        This dir path can then be used to find the next level of information
        in the directory by appending the appropriate next dir structures like:
        - <entity_type_name>.json, for the Entity Type object
        - /entities, for the Entities dir
        """
        root_dir = agent_local_path + "/entityTypes"

        entity_type_paths = []

        for entity_type_dir in os.listdir(root_dir):
            entity_type_dir_path = f"{root_dir}/{entity_type_dir}"
            entity_type_paths.append(entity_type_dir_path)

        return entity_type_paths

    @staticmethod
    def build_lang_code_paths(etype: EntityType):
        """Builds dict of lang codes and file locations.

        The language_codes and paths for each file are stored in a dictionary
        inside of the Entity Type dataclass. This dict is accessed later to
        lint each file and provide reporting based on each language code.
        """
        root_dir = etype.dir_path + "/entities"

        for lang_file in os.listdir(root_dir):
            lang_code = lang_file.split(".")[0]
            lang_code_path = f"{root_dir}/{lang_file}"
            etype.entities[lang_code] = {"file_path": lang_code_path}

    @staticmethod
    def gather_entity_type_metadata(etype: EntityType):
        """Extract metadata for Entity Type for later processing."""
        metadata_file = etype.dir_path + f"/{etype.display_name}.json"

        with open(metadata_file, "r", encoding="UTF-8") as etype_file:
            etype.data = json.load(etype_file)
            etype.resource_id = etype.data.get("name", None)
            etype.kind = etype.data.get("kind", None)

            etype_file.close()

    def lint_language_codes(self, etype: EntityType, stats: LintStats):
        """Executes all Entity based linter rules."""

        for lang_code in etype.entities:
            ent_file_path = Common.get_file_based_on_lang_code_filter(
                etype, lang_code, self.lang_code_filter
            )

            if ent_file_path:
                with open(ent_file_path, "r", encoding="UTF-8") as ent_file:
                    data = json.load(ent_file)
                    entities = data.get("entities", None)
                    etype.entities[lang_code]["entities"] = entities

                    stats = self.rules.run_entity_type_rules(
                        etype, lang_code, stats)

                    ent_file.close()

        return stats

    def lint_entities(self, etype: EntityType, stats: LintStats):
        """Lint the Entity files inside of an Entity Type."""
        if "entities" in os.listdir(etype.dir_path):
            self.build_lang_code_paths(etype)
            stats = self.lint_language_codes(etype, stats)

        else:
            pass
            # TODO pmarlow: Add rule for Entity Type missing Entities

        return stats

    def lint_entity_type(self, etype: EntityType, stats: LintStats):
        """Lint a Single Entity Type dir and all subdirectories."""

        etype.display_name = Common.parse_filepath(
            etype.dir_path, "entity_type"
        )

        self.gather_entity_type_metadata(etype)
        stats = self.lint_entities(etype, stats)

        return stats

    def lint_entity_types_directory(self, agent_local_path: str):
        """Linting the Entity Types dir in the JSON Package structure."""
        start_message = f'{"#" * 10} Begin Entity Types Directory Linter'
        self.console.log(start_message)

        stats = LintStats()

        # Create a list of all Flow paths to iter through
        entity_type_paths = self.build_entity_type_path_list(agent_local_path)
        stats.total_entity_types = len(entity_type_paths)

        # linting starts here
        for entity_type_path in entity_type_paths:
            etype = EntityType()
            etype.verbose = self.verbose
            etype.agent_id = self.agent_id
            etype.dir_path = entity_type_path
            etype.naming_pattern = self.naming_conventions.get(
                "entity_type_name", None)
            stats = self.lint_entity_type(etype, stats)

        header = "-" * 20
        rating = Common.calculate_rating(
            stats.total_issues, stats.total_inspected
        )

        end_message = (
            f"\n{header}\n{stats.total_entity_types} Entity Types "
            f"linted. \n{stats.total_issues} issues found out of "
            f"{stats.total_inspected} inspected."
            f"\nYour Agent Entity Types rated at {rating:.2f}/10\n\n"
        )
        self.console.log(end_message)



"""Common Logger for Rules Definitions"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cxlint.resources.types import Resource

class RulesLogger:
    """Common Logger for Rules output."""
    def __init__(
            self,
            console):

        self.console = console

    @staticmethod
    def create_link(resource):
        link = None

        if resource.agent_id and resource.agent_id != "":
            base = "https://dialogflow.cloud.google.com/cx/"

            link_map = {
                "entity_type": f"/entityTypes?id={resource.entity_type_id}",
                "flow": f"/flows/{resource.flow_id}",
                "fulfillment": f"/flows/{resource.flow_id}"
                f"/flow_creation?pageId={resource.page_id}",
                "intent": f"/intents?id={resource.intent_id}",
                "page": f"/flows/{resource.flow_id}"
                f"/flow_creation?pageId={resource.page_id}",
                "test_case": f"/testCases/{resource.test_case_id}",
                "webhook": f"/webhooks/{resource.webhook_id}"
            }

            path = link_map.get(resource.resource_type, None)
            link = base + resource.agent_id + path

        return link

    def generic_logger(
        self, resource: Resource, rule: str, message: str
    ) -> None:
        """Generic Logger for various resources."""
        url = self.create_link(resource)

        link_map = {
            "entity_type": f"[link={url}]"\
                f"{resource.entity_type_display_name}[/link]",
            "flow": f"[link={url}]{resource.flow_display_name}[/link]",
            "fulfillment": f"[link={url}]{resource.flow_display_name} : "\
                f"{resource.page_display_name}[/link]",
            "intent": f"[link={url}]{resource.intent_display_name}[/link]",
            "page": f"[link={url}]{resource.flow_display_name} : "\
                f"{resource.page_display_name}[/link]",
            "test_case": f"[link={url}]"\
                f"{resource.test_case_display_name}[/link]",
            "webhook": f"[link={url}]{resource.webhook_display_name}[/link]",
        }

        final_link = link_map.get(resource.resource_type, None)
        output = f"{rule} : {final_link} {message}"

        self.console.log(output)



"""Pages linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from configparser import ConfigParser
from typing import Dict, Any

from cxlint.common import Common
from cxlint.rules.pages import PageRules
from cxlint.resources.types import Flow, Page, LintStats, FormParameter
from cxlint.resources.routes import Fulfillments


class Pages:
    """Pages linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.config = config
        self.agent_id = Common.load_agent_id(config)
        self.agent_type = Common.load_agent_type(config)
        self.disable_map = Common.load_message_controls(config)
        self.naming_conventions = Common.load_naming_conventions(config)
        self.rules = PageRules(console, self.disable_map)
        self.routes = Fulfillments(verbose, config, console)

    @staticmethod
    def load_naming_conventions(page: Page, styles: Dict[str, str]):
        """Load all Page naming conventions to the current object."""
        page.naming_pattern_generic = styles.get("page_generic_name", None)
        page.naming_pattern_form = styles.get("page_with_form_name", None)
        page.naming_pattern_webhook = styles.get(
            "page_with_webhook_name", None)

        return page

    @staticmethod
    def build_page_path_list(flow_path: str):
        """Builds a list of files, each representing a Page.

        Ex: /path/to/agent/flows/<flow_dir>/pages/<page_name>.json
        """
        pages_path = f"{flow_path}/pages"

        page_paths = []

        for page in os.listdir(pages_path):
            page_file_path = f"{pages_path}/{page}"
            page_paths.append(page_file_path)

        return page_paths

    @staticmethod
    def get_form_parameter_data(param: Dict[str, Any], page: Page):
        fp = FormParameter(page=page)
        fp.display_name = param.get("displayName", None)
        fp.entity_type = param.get("entityType", None)
        fp.required = param.get("required", None)

        fp.fill_behavior = param.get("fillBehavior", None)

        if fp.fill_behavior:
            fp.init_fulfillment = fp.fill_behavior.get(
                "initialPromptFulfillment", None)
            fp.reprompt_handlers = fp.fill_behavior.get(
                "repromptEventHandlers", None)

        fp.advanced_settings = page.form.get("advancedSettings", None)

        if fp.advanced_settings:
            fp.dtmf_settings = fp.advanced_settings.get("dtmfSettings", None)

        return fp

    def lint_form(self, page: Page, stats: LintStats):
        """Lint the Form and sub-resources within it for the Page."""
        parameters = page.form.get("parameters", None)

        if parameters:
            for param in parameters:
                fp = self.get_form_parameter_data(param, page)
                stats = self.routes.lint_reprompt_handlers(fp, stats)

        return stats


    def lint_page(self, page: Page, stats: LintStats):
        """Lint a Single Page file."""
        page.display_name = Common.parse_filepath(page.page_file, "page")
        page.display_name = Common.clean_display_name(page.display_name)

        page.flow.graph.add_node(page.display_name)

        # TODO
        # Page Display Name from Filename contains special characters so it will
        # not match against page display names stored inside the proto objects
        # Need to implement a parser for symbol translation.
        page.flow.all_pages.add(page.display_name)

        with open(page.page_file, "r", encoding="UTF-8") as page_file:
            page.data = json.load(page_file)
            page.verbose = self.verbose
            page.entry = page.data.get("entryFulfillment", None)
            page.events = page.data.get("eventHandlers", None)
            page.form = page.data.get("form", None)
            page.routes = page.data.get("transitionRoutes", None)
            page.route_groups = page.data.get("transitionRouteGroups", None)

            page.resource_id = page.data.get("name", None)
            page.flow.data[page.display_name] = page.resource_id

            # Order of linting is important here
            stats = self.routes.lint_entry(page, stats)
            stats = self.routes.lint_routes(page, stats)
            stats = self.routes.lint_events(page, stats)
            stats = self.lint_form(page, stats)

            if page.route_groups:
                page = self.routes.set_route_group_targets(page)

            stats = self.rules.run_page_rules(page, stats)

            page_file.close()

        return stats

    def lint_pages_directory(self, flow: Flow, stats: LintStats):
        """Linting the Pages dir inside a specific Flow dir.

        Some Flows may not contain Pages, so we check for the existence
        of the directory before traversing
        """
        if "pages" in os.listdir(flow.dir_path):
            page_paths = self.build_page_path_list(flow.dir_path)

            for page_path in page_paths:
                page = Page(flow=flow)
                page.agent_id = flow.agent_id
                page.page_file = page_path

                page = self.load_naming_conventions(
                    page, self.naming_conventions)

                stats.total_pages += 1
                stats = self.lint_page(page, stats)

        return stats



"""Webhook linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from configparser import ConfigParser

from cxlint.common import Common
from cxlint.rules.webhooks import WebhookRules
from cxlint.resources.types import Webhook, LintStats

class Webhooks:
    """Webhook linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.config = config
        self.agent_id = Common.load_agent_id(config)
        self.disable_map = Common.load_message_controls(config)
        self.naming_conventions = Common.load_naming_conventions(config)

        self.rules = WebhookRules(console, self.disable_map)

    @staticmethod
    def build_webhook_path_list(agent_local_path: str):
        """Builds a list of webhook file locations."""
        root_dir = agent_local_path + "/webhooks"

        webhook_paths = []

        for webhook_file in os.listdir(root_dir):
            webhook_file_path = f"{root_dir}/{webhook_file}"
            webhook_paths.append(webhook_file_path)

        return webhook_paths

    @staticmethod
    def get_service_type(webhook: Webhook) -> str:
        """Get the type of Webhook Service that is cofigured."""
        if "genericWebService" in webhook.data:
            webhook.service_type = "Generic Web Service"

        else:
            webhook.service_type = "Other"

        return webhook.service_type

    def lint_webhook(self, webhook: Webhook, stats: LintStats) -> LintStats:
        """Lint a single Webhook file."""

        with open(webhook.dir_path, "r", encoding="UTF-8") as webhook_file:
            webhook.data = json.load(webhook_file)
            webhook.resource_id = webhook.data.get("name", None)
            webhook.display_name = webhook.data.get("displayName", None)
            webhook.service_type = self.get_service_type(webhook)

            timeout_dict = webhook.data.get("timeout", None)
            if timeout_dict:
                webhook.timeout = timeout_dict.get("seconds", None)

            webhook_file.close()

        stats = self.rules.run_webhook_rules(webhook, stats)

        return stats

    def lint_webhooks_directory(self, agent_local_path: str):
        """Linting the top level Webhooks Dir in the JSON Package structure.

        The following files exist under the `webhooks` dir:
        - <webhook-name>.json
        """
        start_message = f'{"#" * 10} Begin Webhooks Directory Linter'
        self.console.log(start_message)

        stats = LintStats()

        # Create a list of all Webhook paths to iter through
        webhook_paths = self.build_webhook_path_list(agent_local_path)

        # Linting Starts Here
        for webhook_path in webhook_paths:
            webhook = Webhook()
            webhook.verbose = self.verbose
            webhook.agent_id = self.agent_id
            webhook.dir_path = webhook_path
            webhook.naming_pattern = self.naming_conventions.get(
                "webhook_name", None)

            stats.total_webhooks += 1
            stats = self.lint_webhook(webhook, stats)

        header = "-" * 20
        rating = Common.calculate_rating(
            stats.total_issues, stats.total_inspected
        )

        end_message = (
            f"\n{header}\n{stats.total_webhooks} Webhooks linted."
            f"\n{stats.total_issues} issues found out of "
            f"{stats.total_inspected} inspected."
            f"\nYour Agent Webhooks rated at {rating:.2f}/10\n\n"
        )
        self.console.log(end_message)



"""Fulfillment routes linter methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from configparser import ConfigParser
from typing import Dict, Any

from cxlint.common import Common
from cxlint.rules.response_messages import ResponseMessageRules
from cxlint.resources.types import Page, Fulfillment, LintStats, FormParameter


class Fulfillments:
    """Fulfillment routes linter methods and functions."""

    def __init__(self, verbose: bool, config: ConfigParser, console):
        self.verbose = verbose
        self.console = console
        self.config = config
        self.disable_map = Common.load_message_controls(config)
        self.agent_id = Common.load_agent_id(config)
        self.agent_type = Common.load_agent_type(config)
        self.rules = ResponseMessageRules(console, self.disable_map)
        self.route_parameters = {}

    @staticmethod
    def check_for_webhook(page: Page, path: Dict[str, Any]):
        """Check the current route for existence of webhook."""
        if "webhook" in path:
            page.has_webhook = True

    @staticmethod
    def check_for_webhook_event_handlers(route: Fulfillment):
        """Check for Webhook Error Event Handler on Page.

        In this method, we're interested in the following conditions:
         - Page is currently flagged w/webhook = True
         - Page HAS NOT been flagged w/having a webhook error handler
         - The trigger MATCHES the pattern 'webhook.error'

        If a Page and its Route meet all the criteria, we'll flip the bit.
        Otherwise, the webhook handler bit will remain False, causing a rule
        flag."""

        if all(
            [
                route.page.has_webhook,
                not route.page.has_webhook_event_handler,
                "webhook.error" in route.trigger,
            ]
        ):
            route.page.has_webhook_event_handler = True

    def collect_transition_route_trigger(self, route):
        """Inspect route and return all Intent/Condition info."""

        trigger = []
        intent_name = None

        if "intent" in route.data:
            trigger.append("intent")
            intent_name = route.data.get("intent", None)

        if "condition" in route.data:
            trigger.append("condition")

        if len(trigger) > 0:
            trigger = "+".join(trigger)

        if self.verbose and intent_name:
            return f"{trigger} : {intent_name}"

        else:
            return trigger

    def get_trigger_info(self, route):
        """Extract trigger info from route based on primary key."""

        if route.fulfillment_type == "event":
            trigger = f"event : {route.data.get('event', None)}"

        if route.fulfillment_type == "reprompt_handler":
            trigger = f"{route.parameter} : event : "\
                f"{route.data.get('event', None)}"

        if route.fulfillment_type == "transition_route":
            intent_condition = self.collect_transition_route_trigger(route)
            trigger = f"route : {intent_condition}"

        return trigger

    def set_route_group_targets(self, page: Page):
        """Determine Route Targets for Route Group routes."""
        current_page = page.display_name

        for route_group in page.route_groups:
            page.flow.graph.add_edge(current_page, route_group)
            page.flow.graph.add_used_node(route_group)

        return page

    def set_route_targets(self, route: Fulfillment):
        """Determine the Route Targets for the specified route.

        Primary function is to build out the graph structure for the
        Flow based on the current page and where the routes are pointing to.
        The graph structure can then be traversed later to determine any errors
        or inconsistencies in design.
        """
        current_page = route.page.display_name

        route.target_flow = route.data.get("targetFlow", None)
        route.target_page = route.data.get("targetPage", None)

        if route.target_page:
            route.page.flow.graph.add_edge(current_page, route.target_page)
            route.page.flow.graph.add_used_node(route.target_page)

        if route.target_flow:
            route.page.flow.graph.add_edge(
                current_page, f"FLOW: {route.target_flow}"
            )
            route.page.flow.graph.add_used_node(f"FLOW: {route.target_flow}")

        return route

    def update_route_parameters(self, route: Fulfillment, item: Dict[str, str]):
        """Update the Route Parameters map based on new info."""
        flow_name = route.page.flow.display_name
        page_name = route.page.display_name

        flow_data = self.route_parameters.get(flow_name, None)
        page_data = None

        if flow_data:
            page_data = flow_data.get(page_name, None)

        # Flow and Page already exists, append to existing list.
        if page_data:
            self.route_parameters[flow_name][page_name].append(item)

        # Flow data exists, but not Page, so only create the Page list.
        elif flow_data and not page_data:
            self.route_parameters[flow_name][page_name] = [item]

        # Neither the Flow or Page data exists, so create it all.
        else:
            self.route_parameters[flow_name] = {page_name: [item]}


    def lint_fulfillment_type(
        self, stats: LintStats, route: Fulfillment, path: object, key: str
    ):
        """Parse through specific fulfillment types and lint."""
        fulfillment_data = path.get(key, None)

        if fulfillment_data:
            for item in fulfillment_data:
                # This is where each message type will exist
                # text, custom payload, etc.

                # TODO pmarlow: create sub-method parsers per type
                if "text" in item:
                    for text in item["text"]["text"]:
                        stats.total_inspected += 1
                        route.text = text

                        stats = self.rules.run_rm_text_rules(route, stats)

                if "parameter" in item:
                    self.update_route_parameters(route, item)

        return stats

    def lint_reprompt_handlers(self, fp: FormParameter, stats: LintStats):
        """Lint for Reprompt Event Handlers inside Form parameters.

        While Reprompt Event Handlers are technically Events, they differ from
        standard Page level Events because they act on the FormParameter data
        structure, not Fulfillment Route data structure as standard Events do.
        """
        if not fp.reprompt_handlers:
            return stats

        for handler in fp.reprompt_handlers:
            route = Fulfillment(page=fp.page, agent_type=self.agent_type)
            route.data = handler
            route.verbose = self.verbose
            route.agent_id = fp.page.agent_id
            route.fulfillment_type = "reprompt_handler"
            route.parameter = fp.display_name
            route.trigger = self.get_trigger_info(route)
            route = self.set_route_targets(route)
            path = route.data.get("triggerFulfillment", None)
            event = route.data.get("event", None)

            if not path and not event:
                continue

            # Flag for Webhook Handler
            self.check_for_webhook(fp.page, path)

            stats = self.lint_fulfillment_type(stats, route, path, "messages")

        return stats

    def lint_events(self, page: Page, stats: LintStats):
        """Parse through all Page Event Handlers and lint."""
        if not page.events:
            return stats

        for route_data in page.events:
            route = Fulfillment(page=page, agent_type=self.agent_type)
            route.data = route_data
            route.agent_id = page.agent_id
            route.fulfillment_type = "event"
            route.trigger = self.get_trigger_info(route)
            route = self.set_route_targets(route)
            path = route.data.get("triggerFulfillment", None)
            event = route.data.get("event", None)

            if not path and not event:
                continue

            # Flag for Webhook Handler
            self.check_for_webhook_event_handlers(route)

            stats = self.lint_fulfillment_type(stats, route, path, "messages")

        return stats

    def lint_routes(self, page: Page, stats: LintStats):
        """Parse through all Transition Routes and lint."""
        tf_key = "triggerFulfillment"

        if not page.routes:
            return stats

        for route_data in page.routes:
            route = Fulfillment(page=page)
            route.data = route_data
            route.agent_id = page.agent_id
            route.fulfillment_type = "transition_route"
            route.trigger = self.get_trigger_info(route)
            route = self.set_route_targets(route)

            path = route.data.get(tf_key, None)

            if not path:
                continue

            # Flag for Webhook Handler
            self.check_for_webhook(page, path)

            stats = self.lint_fulfillment_type(stats, route, path, "messages")

            # Preset Params can be linted here
            stats = self.lint_fulfillment_type(
                stats, route, path, "setParameterActions"
            )

        return stats

    def lint_entry(self, page: Page, stats: LintStats):
        """Lint Entry Fulfillment on a single page file.

        The Entry Fulfillment to a Page only has 1 "route" (i.e. itself) so
        there is no need to loop through multiple routes, as they don't
        exist for Entry Fulfillment.
        """

        if not page.entry:
            return stats

        route = Fulfillment(page=page)
        route.data = page.entry
        route.agent_id = page.agent_id
        route.fulfillment_type = "entry"
        route.trigger = "entry"
        path = route.data

        self.check_for_webhook(page, path)

        stats = self.lint_fulfillment_type(stats, route, path, "messages")

        return stats
